from typing import overload
from .data_tree import *

@overload
def merge_tree(merge_to: dict, merge_from: dict, children_key: str = 'children', id_key: str = 'id', remove_addtional: bool = True):
    pass


@overload
def merge_tree(merge_to: list, merge_from: list, children_key: str = 'children', id_key: str = 'id', remove_addtional: bool = True):
    pass


def merge_tree(merge_to, merge_from, children_key: str = 'children', id_key: str = 'id', remove_addtional: bool = True):
    # print('merge_to', merge_to, 'merge_from', merge_from)
    if merge_to is None or merge_from is None:
        return
    if not isinstance(merge_to, type(merge_from)):
        raise Exception(
            f'merge_to\'s type({type(merge_to)}) should equal to merge_from\'s({type(merge_from)}).')
    if not isinstance(merge_to, list) and not isinstance(merge_to, dict):
        raise Exception('obj should be list or dict')
    if isinstance(merge_to, list):
        # 收集目标
        merge_to_dict = {}
        for x in merge_to:
            merge_to_dict[x[id_key]] = x

        # 收集来源
        merge_from_dict = {}
        for x in merge_from:
            merge_from_dict[x[id_key]] = x

        # 找到目标原有的项并更新
        additional_merge_to_dict = {}
        for x in merge_to_dict:
            if not x in merge_from_dict:
                additional_merge_to_dict[x] = True  # merge_to 多出的
                continue
            item = merge_from_dict[x]
            merge_tree(merge_to_dict[x], item,
                       children_key, id_key, remove_addtional)
            del merge_from_dict[x]

        # 删除原来多余的项
        if remove_addtional:
            index_to_remove = []
            index_counter = 0  # 每次删除序号向前
            for index, x in enumerate(merge_to):
                if x[id_key] in additional_merge_to_dict:
                    index_to_remove.append(index-index_counter)
                    index_counter += 1
            for x in index_to_remove:
                merge_to.pop(x)

        # 找到来源多余的项附加
        for x in merge_from_dict:
            item = merge_from_dict[x]
            merge_to.append(item)
        return

    # 目标与来源同级
    if merge_to[id_key] != merge_from[id_key]:
        raise Exception('merge_to\'s id should equal to merge_from\'s.')

    # 将本级数据更新
    raw_children = merge_to.get(children_key)
    merge_to.update(merge_from)
    merge_to[children_key] = raw_children

    # 更新子层级数据
    from_children = merge_from.get(children_key)
    if from_children and len(from_children):
        if raw_children is None:
            merge_to[children_key] = []
        merge_tree(merge_to.get(children_key), from_children)
    return merge_to



from __future__ import annotations
import json
import redis
from functools import wraps
import threading
from sgtpyutils.logger import logger
from sgtpyutils.functools import AssignableArg
import time


class RedisCallbackEvent:
    data: str = None
    channel: str = None

    def __init__(self, channel: str, data: str) -> None:
        self.channel = channel
        self.data = data

    def __str__(self) -> str:
        return f'[{self.channel}]{self.data}'


Subscribers: dict[str, list[function]] = {}


class SubscriberWorker(threading.Thread):
    pb: redis.client.PubSub = None
    client: redis.client.Redis = None

    DEFAULT_Config = {
        'host': '127.0.0.1',
        'port': 38379,
        'db': 7,
        'password': 'defaultredispsw@ts.8.8'
    }

    @staticmethod
    def get_client():
        if SubscriberWorker.client is None:
            config = SubscriberWorker.DEFAULT_Config
            SubscriberWorker.client = redis.client.Redis(**config)

        return SubscriberWorker.client

    @staticmethod
    def get_pubsub():
        if SubscriberWorker.pb is None:
            SubscriberWorker.pb = SubscriberWorker.get_client().pubsub()
        return SubscriberWorker.pb

    def __init__(self) -> None:
        super().__init__(name=__class__.__name__,  daemon=True)

    def handle_msg(self, channel: str, value: str):
        callbacks = Subscribers.get(channel)
        if not callbacks:
            logger.debug(f'redis_subscribe[{channel}]:{value}')
            return
        for callback in callbacks:
            arg = AssignableArg(method=callback)
            data = RedisCallbackEvent(channel, value)
            arg.assign_by_type(RedisCallbackEvent, data)
            try:
                callback(*arg.args, **arg.kwargs)
            except Exception as ex:
                msg = 'fail run redis_callback:'
                logger.warning(f'{msg}{callback.__name__},{ex}')

    def run_once(self):
        pb = SubscriberWorker.get_pubsub()
        for x in pb.listen():
            _type = x.get('type')
            if _type == 'subscribe':
                logger.debug(f'new subscribe:{x}')
                continue
            channel = x.get('channel')
            if isinstance(channel, bytes):
                channel = channel.decode()

            data = x.get('data')
            if isinstance(data, bytes):
                data = data.decode()

            self.handle_msg(channel, data)
        logger.debug('redis_subscribe handler run_once completed')

    def run(self) -> None:
        while True:
            self.run_once()
            time.sleep(0.1)

        return super().run()


_redis_callbacker = SubscriberWorker()
_redis_callbacker.start()


class RedisSubscriber:

    @staticmethod
    def publish(event: RedisCallbackEvent):
        channel = event.channel  # 'sec:device:status:update'
        data = event.data
        if isinstance(data, dict):
            data = json.dumps(data)
        if not isinstance(data, str):
            data = str(data)
        logger.debug(f'send message at [{channel}]{data}')
        return SubscriberWorker.get_client().publish(channel, data)

    @staticmethod
    def redis_subscribe_callback(channel: str):
        '''
        当有推送时调用
        method的RedisCallbackData参数将被填充并尝试调用
        '''
        # 如无此句，将导致因闭包channel变量覆盖
        # 从而只有最后一个注册的事件能够被回调
        logger.debug(f'start redis_subscribe hook:{channel}')

        def wrap(method: function):
            return RedisSubscriber.redis_subscribe(channel, method)
        return wrap

    @staticmethod
    def redis_subscribe(channel: str, callback_method: function):
        if not Subscribers.get(channel):
            Subscribers[channel] = []
            SubscriberWorker.get_pubsub().subscribe(channel)

        Subscribers[channel].append(callback_method)

        @wraps(callback_method)
        def wrapper(*args, **kwargs):
            return callback_method(*args, **kwargs)
        return wrapper

    @staticmethod
    def require_redis():
        def require_func(method: function):
            @wraps(method)
            def wrapper(*args, **kwargs):
                x_args = AssignableArg(args, kwargs, method)
                client = SubscriberWorker.get_client()
                x_args.assign_by_type(redis.client.Redis, client)
                result = method(*x_args.args, **x_args.kwargs)
                return result
            return wrapper
        return require_func



import re
import json
import asyncio
import inspect
import importlib
import dataclasses
from pathlib import Path
from contextvars import copy_context
from functools import wraps, partial
from contextlib import asynccontextmanager
from typing_extensions import ParamSpec, get_args, override, get_origin
from typing import (
    Any,
    Type,
    Tuple,
    Union,
    Generic,
    TypeVar,
    Callable,
    Optional,
    Coroutine,
    AsyncGenerator,
    ContextManager,
    overload,
)


P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


def escape_tag(s: str) -> str:
    """用于记录带颜色日志时转义 `<tag>` 类型特殊标签

    参考: [loguru color 标签](https://loguru.readthedocs.io/en/stable/api/logger.html#color)

    参数:
        s: 需要转义的字符串
    """
    return re.sub(r"</?((?:[fb]g\s)?[^<>\s]*)>", r"\\\g<0>", s)


def generic_check_issubclass(
    cls: Any, class_or_tuple: Union[Type[Any], Tuple[Type[Any], ...]]
) -> bool:
    """检查 cls 是否是 class_or_tuple 中的一个类型子类。

    特别的：

    - 如果 cls 是 `typing.Union` 或 `types.UnionType` 类型，
      则会检查其中的所有类型是否是 class_or_tuple 中一个类型的子类或 None。
    - 如果 cls 是 `typing.TypeVar` 类型，
      则会检查其 `__bound__` 或 `__constraints__`
      是否是 class_or_tuple 中一个类型的子类或 None。
    """
    if cls == class_or_tuple:
        return True  # 两者都是type
    if cls.__class__ != type:  # 前者是type
        return type(cls) == type(class_or_tuple)
    try:
        return issubclass(cls, class_or_tuple)
    except TypeError:
        origin = get_origin(cls)
        if origin:
            # avoid class check error (typing.Final, typing.ClassVar, etc...)
            try:
                return issubclass(origin, class_or_tuple)
            except TypeError:
                return False
        elif isinstance(cls, TypeVar):
            if cls.__constraints__:
                return all(generic_check_issubclass(type_, class_or_tuple) for type_ in cls.__constraints__)
            elif cls.__bound__:
                return generic_check_issubclass(cls.__bound__, class_or_tuple)
        return False


def is_coroutine_callable(call: Callable[..., Any]) -> bool:
    """检查 call 是否是一个 callable 协程函数"""
    if inspect.isroutine(call):
        return inspect.iscoroutinefunction(call)
    if inspect.isclass(call):
        return False
    func_ = getattr(call, "__call__", None)
    return inspect.iscoroutinefunction(func_)


def is_gen_callable(call: Callable[..., Any]) -> bool:
    """检查 call 是否是一个生成器函数"""
    if inspect.isgeneratorfunction(call):
        return True
    func_ = getattr(call, "__call__", None)
    return inspect.isgeneratorfunction(func_)


def is_async_gen_callable(call: Callable[..., Any]) -> bool:
    """检查 call 是否是一个异步生成器函数"""
    if inspect.isasyncgenfunction(call):
        return True
    func_ = getattr(call, "__call__", None)
    return inspect.isasyncgenfunction(func_)


def run_sync(call: Callable[P, R]) -> Callable[P, Coroutine[None, None, R]]:
    """一个用于包装 sync function 为 async function 的装饰器

    参数:
        call: 被装饰的同步函数
    """

    @wraps(call)
    async def _wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        loop = asyncio.get_running_loop()
        pfunc = partial(call, *args, **kwargs)
        context = copy_context()
        result = await loop.run_in_executor(None, partial(context.run, pfunc))
        return result

    return _wrapper


@asynccontextmanager
async def run_sync_ctx_manager(
    cm: ContextManager[T],
) -> AsyncGenerator[T, None]:
    """一个用于包装 sync context manager 为 async context manager 的执行函数"""
    try:
        yield await run_sync(cm.__enter__)()
    except Exception as e:
        ok = await run_sync(cm.__exit__)(type(e), e, None)
        if not ok:
            raise e
    else:
        await run_sync(cm.__exit__)(None, None, None)


@overload
async def run_coro_with_catch(
    coro: Coroutine[Any, Any, T],
    exc: Tuple[Type[Exception], ...],
    return_on_err: None = None,
) -> Union[T, None]:
    ...


@overload
async def run_coro_with_catch(
    coro: Coroutine[Any, Any, T],
    exc: Tuple[Type[Exception], ...],
    return_on_err: R,
) -> Union[T, R]:
    ...


async def run_coro_with_catch(
    coro: Coroutine[Any, Any, T],
    exc: Tuple[Type[Exception], ...],
    return_on_err: Optional[R] = None,
) -> Optional[Union[T, R]]:
    """运行协程并当遇到指定异常时返回指定值。

    参数:
        coro: 要运行的协程
        exc: 要捕获的异常
        return_on_err: 当发生异常时返回的值

    返回:
        协程的返回值或发生异常时的指定值
    """

    try:
        return await coro
    except exc:
        return return_on_err


def get_name(obj: Any) -> str:
    """获取对象的名称"""
    if inspect.isfunction(obj) or inspect.isclass(obj):
        return obj.__name__
    return obj.__class__.__name__


def path_to_module_name(path: Path) -> str:
    """转换路径为模块名"""
    rel_path = path.resolve().relative_to(Path.cwd().resolve())
    if rel_path.stem == "__init__":
        return ".".join(rel_path.parts[:-1])
    else:
        return ".".join(rel_path.parts[:-1] + (rel_path.stem,))


def resolve_dot_notation(
    obj_str: str, default_attr: str, default_prefix: Optional[str] = None
) -> Any:
    """解析并导入点分表示法的对象"""
    modulename, _, cls = obj_str.partition(":")
    if default_prefix is not None and modulename.startswith("~"):
        modulename = default_prefix + modulename[1:]
    module = importlib.import_module(modulename)
    if not cls:
        return getattr(module, default_attr)
    instance = module
    for attr_str in cls.split("."):
        instance = getattr(instance, attr_str)
    return instance


class classproperty(Generic[T]):
    """类属性装饰器"""

    def __init__(self, func: Callable[[Any], T]) -> None:
        self.func = func

    def __get__(self, instance: Any, owner: Optional[Type[Any]] = None) -> T:
        return self.func(type(instance) if owner is None else owner)



from __future__ import annotations
from sgtpyutils.logger import logger
CONFIG_Dict: dict[str, BaseConfiguration] = {}


class BaseConfiguration:
    def __init__(self, key: str) -> None:
        if CONFIG_Dict.get(key):
            logger.warning(f'duplicated config:{key}')
        CONFIG_Dict[key] = self

        self._cache: dict[str, bytes] = {}

    def get(self, key: str) -> bytes:
        return self._cache.get(key)

    def set(self, key: str, value: bytes):
        self._cache[key] = value




from __future__ import annotations
from functools import wraps
import traceback
from sgtpyutils.datetime import DateTime
import enum
import pathlib2


class ExceptionHandlerOpt(enum.Flag):
    NoAction = 0
    ResumeRun = 1
    LogFile = 2


class ExceptionHandler:

    @staticmethod
    def path_current(x: DateTime):
        return f'./Exceptions/ex_{x.tostring("%Y%m%d")}.log'

    @staticmethod
    def logExInfo(path: str, exception: Exception):
        data = traceback.format_exception(exception)
        f = pathlib2.Path(path)
        f.parent.mkdir(parents=True, exist_ok=True)
        with open(f.absolute().as_posix(), 'a', encoding='utf-8') as out:
            out.write(f'[{DateTime()}]Exception:{data}\n')

    @classmethod
    def handleException(cls, opt: ExceptionHandlerOpt = ExceptionHandlerOpt.ResumeRun | ExceptionHandlerOpt.LogFile, path: callable = path_current):
        def require_func(method: function):
            @wraps(method)
            def wrapper(*args, **kwargs):
                result = None
                try:
                    result = method(*args, **kwargs)
                except Exception as ex:
                    if opt & ExceptionHandlerOpt.LogFile != ExceptionHandlerOpt.NoAction:
                        cls.logExInfo(path(DateTime()), ex)
                    if opt & ExceptionHandlerOpt.ResumeRun != ExceptionHandlerOpt.NoAction:
                        return result
                    raise ex
                return result
            return wrapper
        return require_func



def all_key_to_lower(data: dict, recursive: int = 1e3):
    if data is None:
        return None
    if not isinstance(data, dict):
        return data
    result = {}
    for x in data:
        val = data[x]
        if recursive > 0:
            if isinstance(val, dict):
                val = all_key_to_lower(val, recursive-1)
                
        result[x.lower()] = val

    return result



from sgtpyutils.extensions.itertools import *


def test_iter():
    for index, x in run_cycle('123', length=3):
        pass
    assert index == 27 -1

def test_iter_full():
    for index, x in run_cycle_full('123', length=3):
        pass
    assert index == 3 + 9 + 27 - 1  # 从0开始故-1，每级指数3



import traceback
import threading
import pathlib2
from functools import wraps
import os
import json
from sgtpyutils.datetime import DateTime
from sgtpyutils.functools import AssignableArg
from sgtpyutils.logger import logger
from ..BaseSaver import ISaver


class DateEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, DateTime):
            return obj.tostring()
        return json.JSONEncoder.default(self, obj)


class DatabaseData:
    def __init__(self, data: dict, serializer: callable, deserializer: callable, name: str = None):
        self.data = data
        self.serializer = serializer
        self.deserializer = deserializer
        self.name = name

    def to_dict(self):
        if self.serializer:
            try:
                result = self.serializer(self.data)
                return result
            except Exception as ex:
                logger.error(
                    f'[{self.name}]fail to serializer {self.data} ,ex:{ex}')
                self.data = {}
        return self.data or {}

    def from_dict(self, data: dict):
        if self.deserializer:
            try:
                self.data = self.deserializer(data)
            except Exception as ex:
                logger.error(
                    f'[{self.name}]fail to deserializer {data} ,ex:{ex}')
                self.data = {}
            return
        self.data = data or {}


DBS_DEBUG = True


class Database(ISaver):
    lock = threading.Lock()
    cache: dict[str, DatabaseData] = {}  # 通过DatabaseData[T]存储实现传地址
    # TODO 使用真实的传址
    root_path: str = None
    raw_value: dict
    '''数据原始对象'''
    database: str
    '''数据库名称'''

    def __init__(
            self,     database: str,
            serializer: callable = None, deserializer: callable = None,
            log: bool = True) -> None:
        self.log = log

        self.database = database
        self.data_obj = DatabaseData(
            None, serializer, deserializer, name=database)

        self.load_db()

    def check_file(self):
        return Database.ensure_file(self.database_filename)

    @staticmethod
    def ensure_file(path: str, log: bool = True):
        xxx = pathlib2.Path(path)
        if not xxx.parent.is_dir():
            t_path = xxx.parent.as_posix()
            if log:
                logger.warning(f'target-db-path not exist,create:{t_path}')
            os.makedirs(t_path)
        if not xxx.exists():
            t_file = xxx.as_posix()
            if log:
                logger.warning(f'target-db-file not exist,create:{t_file}')
            with open(t_file, 'w'):
                pass

    def load_db(self, reload: bool = False):
        with Database.lock:
            if not reload:
                if prev := Database.cache.get(self.database):
                    # 覆盖为缓存
                    self.data_obj = prev
                    return self.value

            pre = 'new database cache is loaded:'
            suf = f',id:{hex(id(self.data_obj))}'
            if self.log:
                debug_func = logger.debug if reload else logger.warning
                debug_func(f'{pre}{self.database_filename}{suf}')
            self.check_file()
            with open(self.database_filename, 'r', encoding='utf-8') as f:
                data = f.read()
                if len(data) < 3:
                    data = '{}'
                self.data_obj.from_dict(json.loads(data))

            Database.cache[self.database] = self.data_obj
            return self.value

    def save(self):
        if hasattr(self, 'is_deleted'):
            return False
        if Database.cache.get(self.database) is None:
            return  # 已被删除
        data = self.data_obj.to_dict()
        return Database.save_direct(self.database_filename, data)

    @staticmethod
    def save_direct(path: str, data: dict):
        try:
            if data is not None:
                x = isinstance(data, dict) or isinstance(data, list)
                assert x, f'无效的类型:{type(data)}{data}'

            Database.ensure_file(path)
            Database.ensure_key_is_str(path, data)
            with open(path, 'w', encoding='utf-8') as f:
                # print(f'{path}:type{type(data)},keys:{len(data)},')
                data_raw = json.dumps(
                    data, cls=DateEncoder, ensure_ascii=False)
                if DBS_DEBUG:
                    print('save_direct', path, len(data_raw))
                f.write(data_raw)
        except Exception as ex:
            logger.error(f'file-db:fail on save_direct {path}： {ex}')

    @staticmethod
    def ensure_key_is_str(name: str, data: dict):
        if isinstance(data, dict):
            for x in data:
                if isinstance(x, str):
                    continue
                logger.warning(f'{name},key [{x}] invalid type:{type(x)}')
                # 此处不要尝试修复，因为可能会有重复key导致数据丢失

    @property
    def database_filename(self) -> str:
        '''文件完整路径'''
        return Database._database_filename(self.database, self.base_path)

    def _database_filename(data: str, base: str = None) -> str:
        if base is None:
            base = Database.get_base_path(data)
        return f'{base}{data}.json'

    @property
    def base_path(self):
        if hasattr(self, '_basepath'):
            return self._basepath
        self._basepath = Database.get_base_path(self.database)
        return self._basepath

    @staticmethod
    def get_base_path(database: str):
        base_path: str
        '''基础路径，当使用相对路径时使用'''
        is_abs: bool
        '''是否是绝对路径，影响base_path'''

        is_abs = pathlib2.Path(database).is_absolute()
        if is_abs:
            base_path = ''
        else:
            current = pathlib2.Path(__file__).parent.parent.parent
            root_path = current.joinpath('database')
            base_path = f'{root_path}{os.sep}'
        return base_path

    @property
    def value(self) -> dict:
        return Database.cache.get(self.database).data

    @value.setter
    def value(self, value: any) -> dict:
        if not Database.cache.get(self.database):
            Database.cache[self.database] = self.data_obj
        Database.cache[self.database].data = value

    def get(self, key: str, default: any = None):
        if key not in self.value:
            self.value[key] = default
        return self.value.get(key)

    @staticmethod
    def require_database(db_path: str):
        def func(target_mathod: callable):
            @wraps(target_mathod)
            def wrapper(*args, **kwargs):
                db = Database(db_path)
                arg = AssignableArg(args, kwargs, target_mathod)
                arg.assign_by_type(Database, db)
                return target_mathod(*arg.args, **arg.kwargs)
            return wrapper
        return func

    @staticmethod
    def save_all():
        stacks = str.join('\n', [str(x) for x in traceback.extract_stack()])
        print(f'attempt to save all from {stacks}')
        for x in Database.cache:
            data_obj = Database.cache[x]
            path = Database._database_filename(x)
            Database.save_direct(path, data_obj.to_dict())

    def delete(self):
        logger.warning(f'attempt to delete db:{self.database}')
        self.is_deleted = True
        if Database.cache.get(self.database) is not None:
            del Database.cache[self.database]

        path = self.database_filename
        if not os.path.exists(path):
            return
        os.remove(path)


__x = pathlib2.Path(__file__)
__x = __x.parent.parent.parent.joinpath('database')
Database.root_path = __x
# should let user manually run instead of auto
# because here shall cause data-lost
# why? seeems register let static-class to be a new instances.
# atexit.register(Database.save_all)



import pathlib2
from sgtpyutils.database import filebase_database

import os


def test_db():
    db = filebase_database.Database('test')
    assert os.path.exists(db.database_filename)

    db.value['test'] = 'test'
    assert db.value['test'] == 'test'

    db.save()

    db = filebase_database.Database('test')
    assert db.value['test'] == 'test'

    db.delete()


def test_aop():
    @filebase_database.Database.require_database('test_aop')
    def test(database: filebase_database.Database):
        assert database
        assert database.database_filename
        database.delete()
    test()


def test_save_all():
    for i in range(10):
        db_name = f'test_save_all_{i}'
        db = filebase_database.Database(db_name)
        db.value['x'] = i

    filebase_database.Database.save_all()

    for i in range(10):
        db_name = f'test_save_all_{i}'
        db = filebase_database.Database(db_name)
        assert db.value['x'] == i

    for i in range(10):
        db_name = f'test_save_all_{i}'
        db = filebase_database.Database(db_name)
        db.delete()
        assert not os.path.exists(db.database_filename)


def test_multi_edit():
    db_name = 'test_multi_edit'
    db = filebase_database.Database(db_name)
    db.value['common'] = 1

    db2 = filebase_database.Database(db_name)
    db2.value['common'] += 1

    db3 = filebase_database.Database(db_name)
    db3.value['common2'] = '123'

    db_check = filebase_database.Database(db_name)
    assert db_check.value['common'] == 2
    assert db_check.value['common2'] == '123'

    db_check.delete()


def test_serializer():
    class TestSerializer:
        def __init__(self, data: dict = None) -> None:
            if data is None:
                return
            self.load_data(data)

        def to_dict(self) -> dict:
            return {
                'name': self.name
            }

        @classmethod
        def from_dict(cls, data: dict):
            target = TestSerializer()
            return target.load_data(data)

        def load_data(self, data: dict):
            if data is None:
                data = {}
            self.name = data.get('name')
            return self

    db_name = 'test_serializer'
    serialer = TestSerializer.to_dict
    def deserialer(x): return TestSerializer.from_dict(x)
    db = filebase_database.Database(db_name, serialer, deserialer)

    obj = TestSerializer({'name': 'xxx'})
    db.value = obj
    db.save()

    filebase_database.Database.cache = {}

    db = filebase_database.Database(db_name, serialer, deserialer)
    data: TestSerializer = db.value
    assert data.name == 'xxx'

    data.name = 'xx2'
    db.save()

    filebase_database.Database.cache = {}

    db = filebase_database.Database(db_name, serialer, deserialer)
    data: TestSerializer = db.value
    assert data.name == 'xx2'

    filebase_database.Database.save_all()
    db.delete()


def test_global_independent():
    path = pathlib2.Path(__file__).parent
    db_name = f'{path}test_global_independent'
    db = filebase_database.Database(db_name)
    db1 = filebase_database.Database(db_name)
    assert id(db.data_obj) == id(db1.data_obj)
    db1.value['test'] = '1'

    db_name2 = 'test_global_independent2'
    db2 = filebase_database.Database(db_name2)
    assert id(db.data_obj) != id(db2.data_obj)

    db2.value['test'] = '2'
    db2.value['test2'] = '2.1'

    assert db1.value.get('test') == '1'
    assert db2.value.get('test') == '2'
    assert db1.value.get('test2') is None
    assert db2.value.get('test2') == '2.1'

    db1.delete()
    db2.delete()



from typing import Literal


def padding_to(raw: bytes, padding: int, max_length: int = None, padding_char: bytes = b'\0', align: Literal['left', 'right'] = 'right'):
    '''
    sometimes crypto need padding length to its requires
    '''
    c = len(raw)/padding
    block: int = int(c)
    if block != c:
        block += 1
    func = raw.rjust if align == 'right' else raw.ljust
    result = func(padding*block, padding_char)
    if max_length:
        result = result[0:max_length]
    return result



from sgtpyutils.math import *


def test_sci_num_raw():
    value = 12345678
    result = number_sci(value)
    assert result == '12M345K678'

    result = number_sci_raw_value(value)
    assert result == '12.345M'




import redis
from .BaseConfiguration import *


class RedisConfiguration(BaseConfiguration):
    def __init__(self, key: str, **kwargs) -> None:
        self.client = redis.StrictRedis(**kwargs)
        super().__init__(key)

    def set(self, key: str, value: bytes):
        return self.client.set(key, value)

    def get(self, key: str):
        return self.client.get(key)



from typing import Dict, Any, AnyStr
from sgtpyutils.functools import AssignableArg


class TargetType:
    def __init__(self, value) -> None:
        self.value = value


def require_by_type(method):
    def wrapper(*args, **kwargs):
        arg = AssignableArg(args, kwargs, method)
        arg.assign_by_type(TargetType, TargetType(1))
        arg.assign_by_type(int, 2)
        arg.assign_by_type(list[int], [3])
        return method(*arg.args, **arg.kwargs)
    return wrapper


@require_by_type
def target_func1(testInt: int, testType: TargetType, x: list[int]):
    return testInt, testType, x


def test_auto_assign_by_type():
    result = target_func1()
    assert result[0] == 2
    assert result[1].value == 1
    assert result[2][0] == 3


def require_by_name(method):
    def wrapper(*args, **kwargs):
        arg = AssignableArg(args, kwargs, method)
        arg.assign_if_not_assigned('assigned', 11)
        arg.assign_if_not_assigned('not_assigned', 12)
        arg.assign_if_not_exist('kw_arg', 13)
        return method(*arg.args, **arg.kwargs)
    return wrapper


@require_by_name
def target_func2(assigned: int, not_assigned: int, **kwargs):
    return assigned, not_assigned, kwargs


def test_auto_assign_by_name():
    assigned, not_assigned, kwargs = target_func2(1)
    assert assigned == 1
    assert not_assigned == 12
    assert kwargs.get('kw_arg') == 13


class ATypeClass:
    def __init__(self, a) -> None:
        self.a = a


AType = ATypeClass

xDict = Dict[Any, Any]
xStr = str


def require_by_name(method):
    def wrapper(*args, **kwargs):
        arg = AssignableArg(args, kwargs, method)
        current_value = arg.get_assign_by_type(xDict)
        assert current_value['v'] == 3

        current_value = arg.get_assign_by_type(xStr)
        assert current_value == '2'

        current_value = arg.get_assign_by_type(list[int])
        assert current_value == [5]

        return method(*arg.args, **arg.kwargs)
    return wrapper


@require_by_name
def func_to_assign(a: AType, b: xStr, c: xDict, d: Dict, e: list[int]):
    return a


def test_get_assign():
    func_to_assign(AType(1), '2', {'v': 3}, {'v': 4}, [5])



from typing import overload
import xml.etree.ElementTree as ET
XML_KEY_Attrs = '@attributes'
XML_KEY_Value = 'value'


class XmlParser:
    DESC_Version = '<?xml version: "1.0" encoding="UTF-8"?>'
    DESC_Xmlns = 'http://www.isapi.org/ver20/XMLSchema'
    ATTR_Utf8 = {'encoding': 'UTF-8'}
    ATTR_Xmlns = {'xmlns': DESC_Xmlns, 'version': '2.0'}

    @staticmethod
    @overload
    def dict2xml(data: dict, parent_tag: str = None) -> ET.Element:
        ...

    @staticmethod
    @overload
    def dict2xml(data: list, parent_tag: str = None) -> ET.Element:
        ...

    @staticmethod
    def dict2xml(data: dict, tag: str = None) -> ET.Element:
        '''
        @param tag:str:根层级名称
        @param data:dict:字典
        '''

        result = ET.Element(tag)
        if isinstance(data, list):
            for x in data:
                result.append(XmlParser.dict2xml(x))
            return result

        if not isinstance(data, dict):
            result.text = str(data)
            return result

        for x in data:
            if x == XML_KEY_Attrs:
                result.attrib = data[x]
                continue
            val = data[x]
            result.append(XmlParser.dict2xml(val, x))

        return result

    @staticmethod
    def dict2xmlstr(data: dict) -> str:
        x = XmlParser.dict2xml(data)
        return ET.tostring(x).decode()

    @staticmethod
    def element2dict(element: ET.Element) -> tuple[str, dict]:
        '''
        @return tuple[
            str:tag_name
            dict:详细字典
        ]
        '''
        index = -1
        try:
            # 去掉前缀（如果有的话）
            index = element.tag.index('}')
        except:
            pass
        element.tag = element.tag[index+1:]

        d = {}
        has_attrs = element.attrib
        if has_attrs:
            d[XML_KEY_Attrs] = has_attrs

        children = [x for x in element]
        if len(children) == 0:
            if not has_attrs:
                return element.text
            d[XML_KEY_Value] = element.text
            return d

        for child in children:
            child_data = XmlParser.element2dict(child)
            child_tag = child.tag
            if not child_tag in d:
                d[child_tag] = child_data
                continue
            prev_value = d[child_tag]

            if isinstance(prev_value, list):
                d[child_tag].append(child_data)
                continue
            # 首次出现多项 则转为列表
            d[child_tag] = [prev_value, child_data]

        return d

    @staticmethod
    @overload
    def xml2dict(element: str) -> dict:
        ...

    @staticmethod
    @overload
    def xml2dict(element: ET.Element) -> dict:
        ...

    @staticmethod
    def xml2dict(element) -> dict:
        if isinstance(element, str):
            if len(element) < 2:
                return {}
            element = ET.fromstring(element)

        return XmlParser.element2dict(element) or {}



from __future__ import annotations
import inspect


class AssignableArg:
    def __init__(self, args: tuple = None, kwargs: dict = None, method: function = None) -> None:
        self.args = args or ()
        self.kwargs = kwargs or {}
        self.method = method

    @staticmethod
    def get_method_dict(method: function):
        annotations = enumerate(method.__annotations__)
        args_sets = dict([[x, index] for index, x in annotations])
        return args_sets

    @property
    def method_dict(self):
        return AssignableArg.get_method_dict(self.method)

    def set_args(self, index: int, value: any, method: function = None):
        method = self.method if method is None else method

        preset = self.args[0:index]
        need_append = index - len(preset)  # 当前面的参数没有值时，使用None填充
        if need_append > 0:
            preset = list(preset) + [None] * need_append
            preset = tuple(preset)
        suffix = self.args[index+1:]
        preset = self.check_preset(preset)  # 检查当前kwargs
        self.args = preset + tuple([value]) + suffix

    def check_preset(self, preset: list, method: function = None):
        '''
        检查前缀是否已经有值，有则需要将kwargs中删除
        '''
        method = self.method if method is None else method
        preset_len = len(preset)
        preset_list = list(preset)
        method_dict = self.get_method_dict(method)
        is_updated = False

        keys = [x for x in method_dict]
        for x in keys:
            index = method_dict[x]
            if index < preset_len and preset_list[index] is None:
                # 对前缀且未赋值的实施
                if x in self.kwargs:
                    # 如果kw中有则覆盖，否则使用默认值
                    preset_list[index] = self.kwargs[x]
                    del self.kwargs[x]
                else:
                    preset_list[index] = self.assign_default_value(x, method)
                    pass
                is_updated = True
        if not is_updated:
            return preset
        self.args = tuple(preset_list)
        return self.args

    def assign_default_value(self, key: str, method: function = None):
        '''get argument default-value'''
        method = self.method if method is None else method
        params = inspect.signature(method).parameters
        p_dict = dict([[x, params[x]] for x in params])
        return p_dict[key].default

    def check_if_exist(self, key: str, method: function = None) -> tuple[any, int]:
        '''
        检查函数入参是否包含变量
        包含则返回该变量当前的值及序号，否则返回None和-1
        '''
        method = self.method if method is None else method

        r = self.kwargs.get(key)
        if r:
            return (r, -1)
        arg_index = self.get_method_dict(method).get(key)
        if not arg_index is None:

            # 判断当前变量已被包含，否则使用其默认值
            if len(self.args) > arg_index:
                v = self.args[arg_index]
            else:
                v = self.assign_default_value(key, method)
            return (v, arg_index)

        return (None, -1)

    def assign_if_not_assigned(self, key: str, value: any, method: function = None) -> bool:
        '''
        检查函数入参是否包含变量且非默认值，如不包含则赋值到kwargs
        @return bool:是否赋值
        '''
        # 已有默认值 或 已传入值 则不赋值
        return self._assign_if_empty(key, value, method, True)

    @staticmethod
    def is_empty(value):
        if value is None:
            return True
        if not generic_check_issubclass(type(value), type):
            return False
        return generic_check_issubclass(value, inspect._empty)

    def assign_if_not_exist(self, key: str, value: any, method: function = None) -> bool:
        '''
        检查函数入参是否包含变量，如不包含则赋值到kwargs
        @return bool:是否赋值
        '''
        return self._assign_if_empty(key, value, method, False)

    def _assign_if_empty(self, key: str, value: any, method: function = None, check_arg_index: bool = False):
        method = self.method if method is None else method
        
        _value, _arg_index = self.check_if_exist(key, method)
        if check_arg_index and _arg_index == -1:
            return True
        if not AssignableArg.is_empty(_value):  # empty类型可继续赋值
            return True  # 已有值 不再赋值
        if _arg_index > -1:
            return self.set_args(_arg_index, value, method)
        self.kwargs[key] = value
        return True

    def assign_by_type(self,  target_type: type, value: any, method: function = None):
        method = self.method if method is None else method

        uargs = method.__annotations__
        for x in uargs:
            if not generic_check_issubclass(uargs[x], target_type):
                continue  # not taraget type
            self.assign_if_not_exist(x, value)
        return self

    def get_assign_by_type(self, target_type: type, method: function = None):
        method = self.method if method is None else method

        uargs = method.__annotations__
        for index, x in enumerate(uargs):
            if not generic_check_issubclass(uargs[x], target_type):
                continue  # not taraget type
            over = len(self.args) > index
            user = self.args[index] if over else self.kwargs.get(x)
            user = user or self.kwargs.get(x)  # prevent user from empty
            if user:
                return user
        return None

from .utils import *

def check_if_exist(key: str, args: AssignableArg) -> tuple[any, int]:
    return args.check_if_exist(key)


def assign_if_not_exist(key: str, value: any, args: AssignableArg):
    return args.assign_if_not_exist(key, value)


def assign_by_type(args: tuple, kwargs: dict, method: function, target_type: type, value: any):
    x_args = AssignableArg(args, kwargs, method)
    return x_args.assign_by_type(target_type, value)


def get_assign_by_type(args: tuple, kwargs: dict, method: function, target_type: type):
    x_args = AssignableArg(args, kwargs, method)
    return x_args.get_assign_by_type(target_type)



SCALE_Kilo: tuple[str, str] = ('KMGTPEZY', '33333333')
SCAKE_CN: tuple[str, str] = ('万亿兆京垓秭壤沟涧正载', '44444444444')
SCAKE_Coin: tuple[str, str] = ('铜银金砖', '0222')


def __number_sci_raw(raw: int, scales: tuple[str, str] = None) -> list[tuple[int, str]]:
    '''
    将数值转换为计数
    @param raw int:原始数值
    @param scales tuple[str,str]:转换规则
    @return list:结果
    '''
    if not scales:
        scales = SCALE_Kilo
    current_level = 0
    scale_str = scales[0]
    scale_rate = scales[1]
    scale_rate_max = len(scale_rate)
    result = []
    while True:
        if scale_rate_max <= current_level:
            break
        cur_rank = int(scale_rate[current_level])
        current_limit = pow(10, cur_rank)

        current_str = scale_str[current_level -
                                1] if current_level > 0 else ''  # 当前计量单位
        cur_result = (int(raw % current_limit), current_str)   # 当前数值
        result.append(cur_result)

        raw = int(raw / current_limit)  # 取下一层级数值
        if raw == 0:
            break
        current_level += 1

    result.reverse()
    return result


def number_sci_raw_value(raw: int, scales: tuple[str, str] = None, limit: int = 3) -> str:
    '''
    将数值转换为带小数点的计数
    '''
    r = __number_sci_raw(raw, scales)
    decimal = ''
    index = 1
    max_scale = len(r)
    if max_scale == 0:
        return '0'
    while len(decimal) < limit:
        if index >= max_scale:
            break
        decimal += f'{r[index][0]}'
        index += 1
    decimal = decimal[0:limit]
    result = f'{r[0][0]}.{decimal}{r[0][1]}'
    return result

def number_sci_raw_str(raw: int, scales: tuple[str, str] = None)->str:
    '''
    将数值转换为各个计量单位值
    '''
    r = __number_sci_raw(raw, scales)
    result = [f'{x[0]}{x[1]}' for x in r]
    return result


def number_sci(raw: int, scales: tuple = None) -> str:
    '''
    将数值转换为计数
    @param raw int:原始数值
    @param scales tuple[str,str]:转换规则
    @return str:结果
    '''
    result = number_sci_raw_str(raw, scales)
    return str.join('', result)



import random
from sgtpyutils.logger import logger
import httpx
from .UserAgent import *
ua_generator = FakeUserAgent(browsers=['chrome'], os='linux')
user_agent_pool = [
    ua_generator.random
]


def __get_default_header_arg(**kwargs):
    headers = kwargs.get('headers') or {}  # 设置默认headers
    headers['User-Agent'] = headers.get('User-Agent') or random.choice(
        user_agent_pool)
    kwargs['headers'] = headers
    return kwargs


def get_default_args(**kwargs):
    kwargs['timeout'] = kwargs.get('timeout') or 5
    kwargs = __get_default_header_arg(**kwargs)
    return kwargs


async def send_with_async(method: str, url: str, proxy: dict = None, **kwargs) -> httpx.Response:
    '''
    @param method str:请求方式，不区分大小写 如GET POST PUT PATCH OPTION DELETE HEADER TRACE等
    @param url str:请求的地址
    @param proxy dict:使用的代理 格式为{"http":"http://xxxx:xxx","https":...}
    @param **kwargs:其他参数如 headers json data timeout等
    @return httpx.Response:响应结果
    以指定方式发出请求，并返回请求结果的Response对象
    '''
    kwargs = get_default_args(**kwargs)
    max_try_time = 3
    while True:
        try:
            async with httpx.AsyncClient(proxies=proxy, follow_redirects=True, verify=False) as client:
                req = await client.request(method, url, **kwargs)
                return req
        except TimeoutError:
            max_try_time -= 1
            if max_try_time > 0:
                continue
            msg = f"max_try_time(count={max_try_time}) exceeded to request in httpx({method} -> {url})"
            logger.error(msg)
            return None
        except Exception as ex:
            logger.error(
                f"fail to request in httpx({method} -> {url}):[{type(ex).__name__}]{ex}")
            return None


async def get_url(url: str, proxy: dict = None, **kwargs) -> str:
    '''
    以get方式发出请求，并将返回的结果以plaintext方式处理
    '''
    r = await send_with_async('get', url, proxy, **kwargs)
    return r.text if hasattr(r, 'text') else None


async def get_api(url, proxy: dict = None, **kwargs) -> dict:
    '''
    以get方式发出请求，并将返回的结果以json方式处理
    '''
    r = await send_with_async('get', url, proxy, **kwargs)
    return r.json()


async def post_url(url, proxy: dict = None, **kwargs) -> str:
    '''
    以post方式发出请求，data为form-url-encoded,json为application/json
    '''
    r = await send_with_async('post', url, proxy, **kwargs)
    return r.text


async def data_post(url, proxy: dict = None, **kwargs) -> str:
    '''
    功能与post_url一致
    '''
    return await post_url(url, proxy, **kwargs)


async def get_status(url, proxy: dict = None, **kwargs) -> int:
    '''
     以get方式发出请求，获取请求结果的status_code
    '''
    r = await send_with_async('get', url, proxy, **kwargs)
    return r.status_code


async def get_content(url, proxy: dict = None, **kwargs) -> bytes:
    '''
     以get方式发出请求，获取请求结果的直接内容
    '''
    r = await send_with_async('get', url, proxy, **kwargs)
    return r.content
